from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings


class ClientSettings(BaseModel):
    raise_for_status: bool = False
    timeout: int = 300
    auto_decompress: bool = True


class AppSettings(BaseModel):
    show_Read: bool = False
    mark_as_read_on_open: bool = True


class Settings(BaseSettings):
    client: ClientSettings = Field(default_factory=ClientSettings)
    app: AppSettings = Field(default_factory=AppSettings)



import asyncio
from pathlib import Path
from rich import print
from rich.console import Console
from sqids import Sqids
from sqlalchemy import create_engine, exc, text
from sqlalchemy.orm import Session
import aiohttp
import click
from lazyfeed.config import Settings
from lazyfeed.db import init_db
from lazyfeed.feeds import fetch_feed_metadata
from lazyfeed.models import Feed
from lazyfeed.opml_utils import export_opml, import_opml
from lazyfeed.repositories import FeedRepository
from lazyfeed.tui import LazyFeedApp


console = Console()
sqids = Sqids(
    alphabet="e69auyponmz7lk5vtgwi1r23hb0d8jq4xfcs",
    min_length=3,
)


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx) -> None:
    """
    A modern RSS feed reader for the terminal.
    """

    ctx.ensure_object(dict)

    # Check app config directory.
    app_dir = Path(click.get_app_dir(app_name="lazyfeed"))
    app_dir.mkdir(parents=True, exist_ok=True)

    ctx.obj["app_dir_path"] = app_dir

    # Set up the SQLite database engine.
    sqlite_url = f"sqlite:///{app_dir / 'lazyfeed.db'}"
    engine = create_engine(sqlite_url)
    init_db(engine)

    ctx.obj["engine"] = engine

    # Load settings.
    ctx.obj["settings"] = Settings()

    # If no subcommand, start the TUI.
    if ctx.invoked_subcommand is None:
        ctx.forward(start_tui)


@cli.command(
    name="tui",
    help="Starts the lazyfeed TUI.",
)
@click.pass_context
def start_tui(ctx) -> None:
    engine = ctx.obj["engine"]
    settings = ctx.obj["settings"]
    with Session(engine) as session:
        app = LazyFeedApp(session, settings)
        app.run()


async def _add_feeds(session: Session, _: Settings, urls: list[str]):
    feed_repository = FeedRepository(session)
    already_saved_urls = [feed.url for feed in feed_repository.get_all()]
    new_urls = [url for url in urls if url not in already_saved_urls]

    if not new_urls:
        console.print("[bold red]ERROR:[/bold red] There are no new urls to check!")
        return
    with console.status(
        "Fetching new feeds... This will only take a moment!",
        spinner="earth",
    ):
        async with aiohttp.ClientSession() as client:
            tasks = [fetch_feed_metadata(client, url) for url in urls]
            results = await asyncio.gather(*tasks, return_exceptions=True)
            for url, result in zip(urls, results):
                if isinstance(result, Exception):
                    console.print(
                        f"[bold red]ERROR:[/bold red] While fetching '{url}' something bad happened!"
                    )
                else:
                    assert isinstance(result, Feed)

                    feed_in_db = feed_repository.add(result)
                    console.print(
                        f"[bold green]SUCCESS:[/bold green] Added RSS feed for '{feed_in_db.title}' ({feed_in_db.link})"
                    )


@cli.command(
    name="add",
    help="Add one or more RSS feeds.",
)
@click.argument("urls", nargs=-1)
@click.pass_context
def add_feed(ctx, urls) -> None:
    engine = ctx.obj["engine"]
    settings = ctx.obj["settings"]
    with Session(engine) as session:
        asyncio.run(_add_feeds(session, settings, urls))


@cli.command(
    name="list",
    help="Print a list with all RSS feeds.",
)
@click.pass_context
def list_feeds(ctx):
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        feed_repository = FeedRepository(session)
        feeds = feed_repository.get_all()

        if not len(feeds):
            console.print(
                "[bold red]ERROR:[/bold red] You need to add some feeds first!"
            )
            return

        for feed in feeds:
            print(
                f"- [bold][ {sqids.encode([feed.id])} ][/] '{feed.title}' ( {feed.link} )"
            )


@cli.command(
    name="delete",
    help="Removes feed.",
)
@click.argument("feed_id")
@click.pass_context
def delete_feed(ctx, feed_id):
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        feed_repository = FeedRepository(session)
        decoded_id = sqids.decode(feed_id)[0]
        feed = feed_repository.delete(decoded_id)
        if not feed:
            console.print(
                f"[bold red]ERROR:[/bold red] No feed found with ID '{feed_id}'."
            )
            return

        console.print(
            f"[bold green]SUCCESS:[/bold green] Feed '{feed.title}' deleted correctly!"
        )


@cli.command(
    name="import",
    help="Import RSS feeds from an OPML file.",
)
@click.argument("input", type=click.File("rb"))
@click.pass_context
def import_feeds(ctx, input) -> None:
    engine = ctx.obj["engine"]
    settings = ctx.obj["settings"]
    with Session(engine) as session:
        urls = import_opml(input)
        asyncio.run(_add_feeds(session, settings, urls))


@cli.command(
    name="export",
    help="Export all RSS feeds to an OPML file.",
)
@click.argument(
    "output",
    type=click.File("wb"),
    default="lazyfeed.opml",
)
@click.pass_context
def export_feeds(ctx, output) -> None:
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        feed_repository = FeedRepository(session)
        feeds = feed_repository.get_all()

        if not len(feeds):
            console.print(
                "[bold red]ERROR:[/bold red] You need to add some feeds first!"
            )
            return

        export_opml(feeds, output)


@cli.command(
    name="vacuum",
    help="Reclaim unused spaced in the database.",
)
@click.pass_context
def vacuum(ctx) -> None:
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        try:
            session.execute(text("VACUUM"))
            session.commit()
            console.print(
                "[bold green]SUCCESS:[/bold green] Unused space has been reclaimed!"
            )
        except exc.SQLAlchemyError as e:
            console.print(f"[bold red]ERROR:[/bold red] Something went wrong: {e}!")



from textual.binding import Binding
from textual.message import Message
from textual.widgets import DataTable

from lazyfeed.confirm_modal import ConfirmModal


class Tabloid(DataTable):
    def __init__(self, *args, **kwargs):
        super().__init__(cursor_type="row", header_height=0, *args, **kwargs)

    BINDINGS = DataTable.BINDINGS + [
        Binding("k", "cursor_up", "Cursor Up", show=False),
        Binding("j", "cursor_down", "Cursor Down", show=False),
        Binding("o", "select_cursor", "Open In Browser", show=False),
        Binding("x", "mark_as_read", "Mark Item As Read", show=False),
        Binding("A", "mark_all_as_read", "Mark All As Read", show=False),
    ]

    def on_mount(self) -> None:
        self.add_columns("Title")

        self.border_title = self.app.TITLE
        self.border_subtitle = "↑/k up · ↓/j down · o open · q quit · ? help"
        self.post_message(self.Ready())

    def action_select_cursor(self) -> None:
        row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
        self.remove_row(row_key)
        self.post_message(self.OpenItem(int(row_key.value)))

    def action_mark_as_read(self) -> None:
        row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
        self.remove_row(row_key)
        self.post_message(self.MarkItemAsRead(int(row_key.value)))

    def action_mark_all_as_read(self) -> None:
        def check_confirmation(response: bool | None) -> None:
            if response:
                self.clear()
                self.post_message(self.MarkAllItemsAsRead())

        self.app.push_screen(
            ConfirmModal("Are you sure that you want to mark all items as read?"),
            check_confirmation,
        )

    class Ready(Message):
        pass

    class OpenItem(Message):
        def __init__(self, post_id: int) -> None:
            super().__init__()
            self.post_id = post_id

    class MarkItemAsRead(Message):
        def __init__(self, post_id: int) -> None:
            super().__init__()
            self.post_id = post_id

    class MarkAllItemsAsRead(Message):
        pass



import asyncio
from pathlib import Path
import aiohttp
import click
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from textual import on, work
from textual.app import App, ComposeResult
from lazyfeed.config import Settings
from lazyfeed.db import init_db
from lazyfeed.feeds import fetch_feed
from lazyfeed.help_modal import HelpModal
from lazyfeed.models import Post
from lazyfeed.tabloid import Tabloid
from lazyfeed.repositories import FeedRepository, PostRepository


class LazyFeedApp(App):
    """
    A Textual based application to read RSS feeds in
    the terminal.
    """

    TITLE = "lazyfeed"
    CSS_PATH = "global.tcss"

    BINDINGS = [
        ("?", "display_help", "Display Help Message"),
        ("q", "quit", "Quit"),
    ]

    def __init__(self, session: Session, _: Settings, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._session = session
        self.feeds_repository = FeedRepository(self._session)
        self.post_repository = PostRepository(self._session)

    def compose(self) -> ComposeResult:
        yield Tabloid()

    def action_display_help(self) -> None:
        self.push_screen(HelpModal())

    def on_mount(self) -> None:
        self.tabloid = self.query_one(Tabloid)

    async def on_quit(self) -> None:
        self._session.flush()
        self.app.exit()

    @on(Tabloid.Ready)
    async def start_fetching(self) -> None:
        self.tabloid.loading = True
        self.load_new_posts()

    @on(Tabloid.OpenItem)
    def open_item(self, message: Tabloid.OpenItem) -> None:
        post_in_db = self.post_repository.get(message.post_id)
        if not post_in_db:
            self.notify(
                "Unable to open the selected item",
                severity="error",
            )
            return

        self.open_url(post_in_db.url)
        self.post_repository.update(message.post_id, read=True)

    @on(Tabloid.MarkItemAsRead)
    def mark_item_as_read(self, message: Tabloid.MarkItemAsRead) -> None:
        self.post_repository.update(message.post_id, read=True)

    @on(Tabloid.MarkAllItemsAsRead)
    def mark_all_items_as_read(self) -> None:
        self.tabloid.loading = True

        try:
            self.post_repository.mark_all_as_read()
            self.notify(
                "All items have been marked as read",
                severity="information",
            )
        except Exception:
            self.notify(
                "Something went wrong while marking all items as read!",
                severity="error",
            )
        finally:
            self.tabloid.loading = False

    @work(exclusive=True)
    async def load_new_posts(self) -> None:
        feeds = self.feeds_repository.get_all()
        if not len(feeds):
            self.notify(
                "You need to add some feeds first!",
                severity="warning",
            )
            self.tabloid.loading = False
            return

        async with aiohttp.ClientSession() as client:
            tasks = [fetch_feed(client, feed) for feed in feeds]
            results = await asyncio.gather(*tasks, return_exceptions=True)

            for feed, result in zip(feeds, results):
                if isinstance(result, Exception):
                    self.notify(
                        f"Something bad happened while fetching '{feed.url}'",
                        severity="error",
                    )
                    continue

                entries, etag = result
                if etag:
                    self.feeds_repository.update(feed.id, etag=etag)

                new_entries = []
                for entry in entries:
                    posts_in_db = self.post_repository.get_by_attributes(url=entry.link)
                    if posts_in_db:
                        continue

                    entry_link = getattr(entry, "link", None)
                    entry_title = getattr(entry, "title", None)
                    entry_summary = getattr(entry, "summary", None)
                    if not entry_link or not entry_title:
                        self.notify(
                            f"Something bad happened while fetching '{entry.title}'",
                            severity="error",
                        )
                        continue

                    new_entries.append(
                        Post(
                            feed=feed,
                            url=entry_link,
                            title=entry_title,
                            summary=entry_summary,
                        )
                    )

                self.post_repository.add_in_batch(new_entries)

        pending_posts = self.post_repository.get_by_attributes(read=False)

        self.tabloid.loading = False
        self.tabloid.focus()

        for post in pending_posts:
            label = f"[bold][{post.feed.title}][/bold] {post.title}"
            self.tabloid.add_row(label, key=f"{post.id}")

        self.notify(f"{len(pending_posts)} new posts!")


if __name__ == "__main__":
    app_dir = Path(click.get_app_dir(app_name="lazyfeed"))
    app_dir.mkdir(parents=True, exist_ok=True)

    sqlite_url = f"sqlite:///{app_dir / 'lazyfeed.db'}"
    engine = create_engine(sqlite_url)
    init_db(engine)

    with Session(engine) as session:
        app = LazyFeedApp(session, Settings())
        app.run()
